# -*- coding: utf-8 -*-

from typing import Dict, List, Union, Any
from pathlib import Path
import os
import abc
import json
from decimal import Decimal
from .io import ensure_parent_dir_exist
from .txt import get_file_encoding
try:
    import ijson
except ImportError as e:
    ijson = None

def _read_json(filepath: Union[str, os.PathLike], jsonl: bool, encoding='utf-8', **kwargs):
    """
    The actual read function.
    """
    encoding = get_file_encoding(filepath, encoding=encoding)
    with open(filepath, encoding=encoding) as f:
        if jsonl:
            return [json.loads(x, **kwargs) for x in f]
        else:
            return json.load(f, **kwargs)

def _is_jsonl(filepath: Union[str, os.PathLike], jsonl: bool = None) -> bool:
    if jsonl is None:
        filepath = Path(filepath)
        jsonl = filepath.suffix.lower() == '.jsonl'
    return jsonl

def read_json(
    filepath: Union[str, os.PathLike],
    jsonl: bool = None,
    encoding: str = 'auto',
    **kwargs
) -> Union[Dict[str, Any], List[Any]]:
    """
    An agent for `json.load()` with some default value.
    """
    jsonl = _is_jsonl(filepath, jsonl)
    try:
        return _read_json(filepath, jsonl=jsonl, encoding=encoding, **kwargs)
    except Exception as e:
        # if failed, try again with different arg `jsonl`
        try:
            return _read_json(filepath, jsonl=not jsonl, encoding=encoding, **kwargs)
        except Exception:
            raise e

def save_json(
    filepath: Union[str, os.PathLike],
    data: Union[Dict[str, Any], List[Any]],
    jsonl: bool = None,
    encoding: str = 'utf-8',
    newline: str = '\n',
    indent: int = 2,
    ensure_ascii: bool = False,
    **kwargs
):
    """
    An agent for `json.dump()` with some default value.
    """
    jsonl = _is_jsonl(filepath, jsonl)
    if jsonl and not isinstance(data, list):
        # data should be a list
        raise ValueError("data should be a list when save as jsonl format")
    ensure_parent_dir_exist(filepath)
    with open(filepath, 'w', encoding=encoding, newline=newline) as f:
        if jsonl:
            for x in data:
                f.write(json.dumps(x, ensure_ascii=ensure_ascii, **kwargs))
                f.write(newline)
        else:
            json.dump(data, f, indent=indent, ensure_ascii=ensure_ascii, **kwargs)

def write_json(
    filepath: Union[str, os.PathLike],
    data: Union[Dict[str, Any], List[Any]],
    jsonl: bool = None,
    encoding: str = 'utf-8',
    newline: str = '\n',
    indent: int = 2,
    ensure_ascii: bool = False,
    **kwargs
):
    save_json(
        filepath=filepath,
        data=data,
        jsonl=jsonl,
        encoding=encoding,
        newline=newline,
        indent=indent,
        ensure_ascii=ensure_ascii,
        **kwargs
    )


class _JsonNode:
    def __init__(self, type: str = '', parent: '_JsonNode' = None):
        self._type = ''
        self._value = None
        self._parent = parent
        if type:
            self.type = type

    def clear(self):
        if self._type == 'map':
            self._value.clear()
        elif self._type == 'array':
            self._value.clear()

    @property
    def parent(self):
        return self._parent

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if self._type:
            raise ValueError('type is already set')
        self._type = value
        if value == 'map':
            self._value = {}
        elif value == 'array':
            self._value = []

    @property
    def value(self):
        if not self._type:
            raise ValueError('type is not set')
        if self._type == 'dummy':
            assert isinstance(self._value, _JsonNode)
            return self._value.value
        if self._type == 'map':
            assert isinstance(self._value, dict)
            return {k: v.value for k, v in self._value.items()}
        if self._type == 'array':
            assert isinstance(self._value, list)
            return [v.value for v in self._value]
        return self._value

    @value.setter
    def value(self, value):
        if not self._type:
            raise ValueError('type is not set')
        if self._type in ['dummy', 'map', 'array']:
            raise RuntimeError('cannot set value for dummy, map, array')
        self._value = value

    def __repr__(self):
        return str(self.value)

    def __str__(self):
        return str(self.value)

class StreamJsonReader(abc.ABC):
    """
    Iterate over a json file.
    """
    def __init__(self, filepath: Union[str, os.PathLike], encoding: str = None, limit: int = float('inf')):
        self.filepath = filepath
        self.encoding = encoding
        self.limit = limit
        self._data_type = ''  # dict or list

    @property
    def data_type(self):
        return self._data_type

    def __iter__(self):
        raise NotImplementedError

class BigJsonReader(StreamJsonReader):
    def __iter__(self):
        with open(self.filepath, 'rb') as f:
            parser = ijson.parse(f)
            dummy = node = _JsonNode('')
            cnt = 0
            for prefix, event, value in parser:
                if event == 'start_map':
                    if node.type == 'array':
                        child = _JsonNode(type='map', parent=node)
                        node._value.append(child)
                        node = child
                    else:
                        node.type = 'map'
                elif event == 'end_map':
                    node = node.parent
                elif event == 'start_array':
                    node.type = 'array'
                elif event == 'end_array':
                    node = node.parent
                elif event == 'map_key':
                    assert node.type == 'map', f"{event} {value} {prefix}"
                    child = _JsonNode(parent=node)
                    node._value[value] = child
                    node = child
                else:
                    assert event in ['null', 'boolean', 'integer', 'double', 'number', 'string']
                    if isinstance(value, Decimal):
                        value = float(value)
                    if node.type == 'array':
                        child = _JsonNode(type=event, parent=node)
                        child.value = value
                        node._value.append(child)
                    else:
                        assert not node.type
                        node.type = event
                        node.value = value
                        node = node.parent
                if node == dummy and event not in ['start_map', 'start_array']:
                    assert node.type in ['map', 'array']
                    if node.type == 'map':
                        value = node.value
                        assert isinstance(value, dict)
                        assert len(value) == 1
                        k, v = list(value.items())[0]
                        self._data_type = 'dict'
                        yield k, v
                        node.clear()
                    elif node.type == 'array':
                        value = node.value
                        assert isinstance(value, list)
                        assert len(value) == 1
                        self._data_type = 'list'
                        yield value[0]
                        node.clear()
                    cnt += 1
                    if cnt >= self.limit:
                        break

class JsonlReader(StreamJsonReader):
    @property
    def data_type(self):
        return 'list'

    def __iter__(self):
        with open(self.filepath, encoding=self.encoding) as f:
            for i, line in enumerate(f, 1):
                yield json.loads(line)
                if i >= self.limit:
                    break

def read_big_json(
    filepath: Union[str, os.PathLike],
    jsonl: bool = None,
    encoding: str = 'auto',
) -> StreamJsonReader:
    jsonl = _is_jsonl(filepath, jsonl)
    encoding = get_file_encoding(filepath, encoding=encoding)
    if jsonl:
        return JsonlReader(filepath, encoding=encoding)
    else:
        if ijson is None:
            raise ImportError('ijson is not installed')
        return BigJsonReader(filepath)




# -*- coding: utf-8 -*-

from typing import (
    Union, List, Any, Iterable,
    Callable, Set, Optional, Tuple,
    Dict, Hashable, Sequence,
)
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

_build_in_na_checkers = {
    'always_na': lambda x: True,
    'never_na': lambda x: False,
    'is_none': lambda x: x is None,
}
_NA_CHECKER_TYPES = Union[Callable[[Any], bool], Literal['always_na', 'never_na', 'is_none']]

_build_in_na_converters = {
    'none': lambda x: None,
    'self': lambda x: x,
    'empty': lambda x: [],
    'single': lambda x: [x],
}
_NA_CONVERTER_TYPES = Union[Callable[[Any], Any], Literal['none', 'self', 'empty', 'single']]

def _get_or_default(value: Any, mapping: Dict[Hashable, Any], default_key: Any) -> Any:
    if value is None:
        return mapping[default_key]
    return mapping.get(value, value)

class ArgValueParser(object):
    @classmethod
    def split_and_parse_strs(cls, strings: Union[Sequence[str], str, None],
                             func: Callable[[str], Any] = None,
                             sep=',', do_trim=True, ignore_blank=True) -> Iterable[Any]:
        """
        split and parse multi string values
        :param strings:  sequence of strings
        :param func:     function to parse single string value
        :param sep:      seperator to split single string
        :param do_trim:  trim every word or not
        :param ignore_blank:    ignore blank or not; in some case, this must be `True`
        """
        if isinstance(strings, str):
            strings = [strings]
        if strings:
            for value in strings:
                for x in value.split(sep):
                    if do_trim:
                        x = x.strip()
                    if not x and ignore_blank:
                        continue
                    yield func(x) if func else x

    @classmethod
    def split_strs_to_set(cls, values: Union[Sequence[str], str, None],
                          func: Callable[[str], Any] = None,
                          sep=',', do_trim=True, ignore_blank=True) -> Optional[Set[Any]]:
        """
        split multi string values as words to a set
        """
        return set(cls.split_and_parse_strs(values, func, sep, do_trim, ignore_blank))

    @classmethod
    def split_strs_to_list(cls, values: Union[Sequence[str], str, None],
                           func: Callable[[str], Any] = None,
                           sep=',', do_trim=True, ignore_blank=True) -> Optional[List[Any]]:
        """
        split multi string values as words to a list
        """
        return list(cls.split_and_parse_strs(values, func, sep, do_trim, ignore_blank))

    @staticmethod
    def bound_collection_if_singleton(value: Any, collection_type: type,
                                      elem_type: Union[type, Tuple[type]] = (),
                                      allowed_type=(list, tuple, set)) -> Any:
        """
        if `value` is a singleton element, bound it with a collection type.
        :param value:       the input value may be bounded
        :param collection_type:     may be `list`, `tuple` or `set`
        :param elem_type:       if `value` is instance of `elem_type` but not instance of `allowed_type`, bound  it.
        :param allowed_type:    if `value` is instance of `allowed_type`, return the `value` unchanged.
        """
        if isinstance(value, allowed_type):
            return value
        if isinstance(value, elem_type):
            return collection_type([value])
        return value

    @classmethod
    def bound_list_if_singleton(cls, value: Any, elem_type: Union[type, Tuple[type]] = (),
                                allowed_type=(list, tuple, set)) -> Any:
        """
        If `value` is a singleton element, bound it to be a `list`.
        See more arg docs in `bound_collection_if_singleton()`.
        """
        return cls.bound_collection_if_singleton(value, collection_type=list,
                                                 elem_type=elem_type, allowed_type=allowed_type)

    @classmethod
    def bound_tuple_if_singleton(cls, value: Any, elem_type: Union[type, Tuple[type]] = (),
                                 allowed_type=(list, tuple, set)) -> Any:
        """
        If `value` is a singleton element, bound it to be a `tuple`.
        See more arg docs in `bound_collection_if_singleton()`.
        """
        return cls.bound_collection_if_singleton(value, collection_type=tuple,
                                                 elem_type=elem_type, allowed_type=allowed_type)

    @classmethod
    def bound_set_if_singleton(cls, value: Any, elem_type: Union[type, Tuple[type]] = (),
                               allowed_type=(list, tuple, set)) -> Any:
        """
        If `value` is a singleton element, bound it to be a `set`.
        See more arg docs in `bound_collection_if_singleton()`.
        """
        return cls.bound_collection_if_singleton(value, collection_type=set,
                                                 elem_type=elem_type, allowed_type=allowed_type)

    @staticmethod
    def ensure_collection(value: Any, expected_type: type, collection_type: Union[type, Tuple[type, ...]],
                          na_checker: _NA_CHECKER_TYPES = None, na_converter: _NA_CONVERTER_TYPES = None) -> Any:
        """
        Ensure the value to be a list, tuple or set.
        :param value:   any type value
        :param expected_type:       expected return type, can be list, tuple or set
        :param collection_type:     other collection type to be convert
        :param na_checker:      check if `value` is na, default is 'is_none'
                                str value means some built-in functions:
                                    always_na:  always treat the value as na
                                    never_na:   never treat the value as na
                                    is_none:    test if the value is `None`
        :param na_converter:    if `value` is na, return output of this function, default is 'self'
                                str value means some built-in functions:
                                    none:   `None`
                                    self:   the value no changed
                                    empty:  an empty list
                                    single: a single value list: `[value]`
        :return:    expected to be an instance of `expected_type`, or be `None` for some condition
        """
        na_checker = _get_or_default(na_checker, _build_in_na_checkers, 'is_none')
        if na_checker(value):
            na_converter = _get_or_default(na_converter, _build_in_na_converters, 'self')
            return na_converter(value)
        if isinstance(value, expected_type):
            return value
        if isinstance(value, collection_type):
            return expected_type(value)
        return expected_type([value])

    @classmethod
    def ensure_list(cls, value: Any, na_checker: _NA_CHECKER_TYPES = None,
                    na_converter: _NA_CONVERTER_TYPES = None) -> Optional[List[Any]]:
        """
        Ensure the value to be a list.
        See more arg docs in `ensure_collection()`.
        """
        return cls.ensure_collection(value, expected_type=list, collection_type=(tuple, set),
                                     na_checker=na_checker, na_converter=na_converter)

    @classmethod
    def ensure_tuple(cls, value: Any, na_checker: _NA_CHECKER_TYPES = None,
                     na_converter: _NA_CONVERTER_TYPES = None) -> Optional[Tuple[Any]]:
        """
        Ensure the value to be a tuple.
        See more arg docs in `ensure_collection()`.
        """
        return cls.ensure_collection(value, expected_type=tuple, collection_type=(list, set),
                                     na_checker=na_checker, na_converter=na_converter)

    @classmethod
    def ensure_set(cls, value: Any, na_checker: _NA_CHECKER_TYPES = None,
                   na_converter: _NA_CONVERTER_TYPES = None) -> Optional[Set[Any]]:
        """
        Ensure the value to be a set.
        See more arg docs in `ensure_collection()`.
        """
        return cls.ensure_collection(value, expected_type=set, collection_type=(list, tuple),
                                     na_checker=na_checker, na_converter=na_converter)




from typing import Union, Literal
import os
import chardet

def detect_text_encoding(raw: bytes, should_rename_legacy=True) -> str:
    return chardet.detect(raw, should_rename_legacy=should_rename_legacy).get('encoding')

def detect_file_encoding(path: Union[str, os.PathLike], size=10000, should_rename_legacy=True) -> str:
    with open(path, 'rb') as f:
        raw = f.read(size)
    return detect_text_encoding(raw, should_rename_legacy=should_rename_legacy)

def get_file_encoding(path: Union[str, os.PathLike], encoding: Union[None, Literal['auto'], str] = None) -> str:
    if encoding == 'auto':
        encoding = detect_file_encoding(path)
    return encoding

def read_txt(path: Union[str, os.PathLike], encoding: Union[None, Literal['auto'], str] = None) -> str:
    if encoding == 'auto':
        with open(path, 'rb') as f:
            raw = f.read()
        encoding = detect_text_encoding(raw)
        return raw.decode(encoding)
    with open(path, 'r', encoding=encoding) as f:
        return f.read()

def save_txt(path: Union[str, os.PathLike], content: str, encoding: str = 'utf-8'):
    with open(path, 'w', encoding=encoding) as f:
        f.write(content)

def write_txt(path: Union[str, os.PathLike], content: str, encoding: str = 'utf-8'):
    save_txt(path=path, content=content, encoding=encoding)



import pandas as pd
from typing import Union, Iterable, Dict, Sequence, Any, List, Tuple
from .dataframe import save_dataframe

def _save_excel(file, df, *args, **kwargs):
    # if df is a list of dataframe, then save each dataframe into a sheet
    if isinstance(df, (list, tuple)) and df and all(isinstance(x, pd.DataFrame) for x in df):
        if 'sheet_name' in kwargs:
            kwargs.pop('sheet_name')
        with pd.ExcelWriter(file) as writer:
            for i, x in enumerate(df, 1):
                save_dataframe(writer, x, *args, sheet_name=f"Sheet{i}", **kwargs)
    elif isinstance(df, dict) and df and all(isinstance(x, pd.DataFrame) for x in df.values()):
        if 'sheet_name' in kwargs:
            kwargs.pop('sheet_name')
        with pd.ExcelWriter(file) as writer:
            for name, x in df.items():
                save_dataframe(writer, x, *args, sheet_name=name, **kwargs)
    else:
        return save_dataframe(file, df, *args, **kwargs)

_FILE_TYPES = Union[str, 'pd.WriteBuffer[bytes]', 'pd.WriteBuffer[str]']
_DATA_TYPES = Union[
    pd.DataFrame, Iterable[Union[pd.Series, Dict[str, Any]]],
    List[pd.DataFrame], Tuple[pd.DataFrame], Dict[str, pd.DataFrame]
]

def save_excel(file: _FILE_TYPES, df: _DATA_TYPES,
               *args, sheet_name='Sheet1',
               header: Union[Sequence[str], bool] = True,
               index=False, index_label=None,
               column_mapper: Union[Dict[str, str], Sequence[str]] = None,
               include_columns: Sequence[str] = None,
               exclude_columns: Sequence[str] = None,
               **kwargs):
    """
    save data into file
    :param file:                where to save the data to
    :param df:                  the data
    :param args:                extra args for df.to_xx()
    :param sheet_name:          `sheet_name` for excel format
    :param header:              `header` for excel format
    :param index:               save index or not, see docs in df.to_csv();
                                if set as str and `index_label` not set, `index_label` will be set as this
    :param index_label:         header for the index when `index` is `True`
    :param column_mapper:       rename columns; if set, columns not list here will be ignored
    :param include_columns:     if set, columns not list here will be ignored
    :param exclude_columns:     if set, columns list here will be ignored
    :param kwargs:              extra kwargs for df.to_xx()
    """
    _save_excel(
        file, df, *args,
        sheet_name=sheet_name,
        header=header,
        index=index,
        index_label=index_label,
        column_mapper=column_mapper,
        include_columns=include_columns,
        exclude_columns=exclude_columns,
        **kwargs
    )

def write_excel(
    file: _FILE_TYPES, df: _DATA_TYPES,
    *args, sheet_name='Sheet1',
    header: Union[Sequence[str], bool] = True,
    index=False, index_label=None,
    column_mapper: Union[Dict[str, str], Sequence[str]] = None,
    include_columns: Sequence[str] = None,
    exclude_columns: Sequence[str] = None,
    **kwargs
):
    save_excel(
        file, df, *args,
        sheet_name=sheet_name,
        header=header,
        index=index,
        index_label=index_label,
        column_mapper=column_mapper,
        include_columns=include_columns,
        exclude_columns=exclude_columns,
        **kwargs
    )



#!/usr/bin/env python
# coding: utf-8

from typing import Dict, Any, Union, Collection, List

def flatten_dict(data: Dict[str, Any], prefix="", joiner=".",
                 exclude: Union[None, str, Collection[str]] = None,
                 frozen: Union[None, str, Collection[str]] = None,
                 empty_as_default=False, empty_value=None,
                 res: Dict[str, Any] = None) -> Dict[str, Any]:
    """
    flatten dict as a flat one layer dict
    :param data:      origin dict
    :param prefix:    prefix for key in the dict
    :param joiner:    join symbol for different layer key
    :param exclude:   prefix to be excluded from result
    :param frozen:    keys not to be flattened
    :param empty_as_default:  should set a default value if value is an empty dict
    :param empty_value:       if `empty_as_default` is `True`, used as the default value for empty dict
    :param res:       the result flat layer dict, create a new one if not given.
    """
    if res is None:
        res = {}
    if isinstance(exclude, str):
        exclude = {exclude}
    if isinstance(frozen, str):
        frozen = {frozen}

    # all keys are start with the prefix, ignore data
    if exclude and prefix in exclude:
        return res

    # all keys in data should be frozen
    if frozen and prefix in frozen:
        for k, v in data.items():
            res[prefix+k] = v
        return res

    for k, v in data.items():
        k = prefix + k

        if exclude and k in exclude:
            # only the key should be excluded
            continue

        if frozen and k in frozen:
            # frozen key, keep it as original value
            res[k] = v
            continue

        if isinstance(v, dict):
            if len(v) == 0:
                # empty dict, set as default value if set
                if empty_as_default:
                    res[k] = empty_value
            else:
                # value is a dict, flatten recursively
                flatten_dict(v, prefix=k+joiner, joiner=joiner, exclude=exclude, frozen=frozen, res=res)
        else:
            # normal value, keep it as original value
            res[k] = v

    return res


def flatten_list(data: List[Any], res: List[Any] = None) -> List[Any]:
    """
    Flatten nested list as a flat on layer list.
    :param data:    a nested list
    :param res:     the result flat layer list, create a new one if not given.
    """
    if res is None:
        res = []

    for x in data:
        if isinstance(x, list):
            # flatten recursively
            flatten_list(x, res)
        else:
            res.append(x)

    return res



# -*- coding: utf-8 -*-

from .io import ensure_parent_dir_exist
from .dataframe import read_dataframe, save_dataframe, extract_dataframe_sample, merge_dataframe_rows, iter_dataframe
from .dataframe import is_empty_text, is_nonempty_text, is_blank_text, is_non_blank_text
from .datetime import format_time, format_date
from .arg import ArgValueParser
from .json import read_json, save_json, write_json, read_big_json
from .txt import (
    get_file_encoding,
    read_txt, save_txt, write_txt,
)
from .process import DataframeProcessor
from .excel import save_excel, write_excel
from .utils import flatten_dict, flatten_list
from .version import __version__

__all__ = [
    'ensure_parent_dir_exist',
    'read_dataframe', 'save_dataframe', 'extract_dataframe_sample', 'merge_dataframe_rows', 'iter_dataframe',
    'is_empty_text', 'is_nonempty_text', 'is_blank_text', 'is_non_blank_text',
    'format_time', 'format_date',
    'ArgValueParser',
    'read_json', 'save_json', 'write_json', 'read_big_json',
    'get_file_encoding',
    'read_txt', 'save_txt', 'write_txt',
    'save_excel', 'write_excel',
    'DataframeProcessor',
    'flatten_dict', 'flatten_list',
    '__version__',
]



# -*- coding: utf-8 -*-

from typing import Union
import pandas as pd
import datetime

def format_time(time: Union[str, int, float, datetime.datetime] = None, fmt='%Y-%m-%d %H:%M:%S') -> str:
    if time is None:
        time = datetime.datetime.now()
    elif isinstance(time, (int, float)):
        time = datetime.datetime.fromtimestamp(time)
    elif isinstance(time, str):
        time = pd.to_datetime(time)
    else:
        if not isinstance(time, datetime.datetime):
            raise ValueError(f"Unexpected type: {type(time)}")
    return time.strftime(fmt)

# when format a date, no sep is used more
def format_date(date: Union[str, int, float, datetime.datetime] = None, sep='') -> str:
    return format_time(date, fmt=sep.join(['%Y', '%m', '%d']))



import abc
import tqdm
import pandas as pd
from typing import (
    Any, Dict, Hashable, List,
    Tuple, Union, Iterable, Optional,
)
from .dataframe import (
    read_dataframe,
    save_dataframe,
)

class BaseProcessor(abc.ABC):
    """
    Base class for processing data.
    """

    @abc.abstractmethod
    def read_single_file(self, filepath: str) -> Any:
        """
        Actual method to read data from a single file.
        """

    def merge_input_data(self, data: Iterable[Any]) -> Any:
        """
        Merge data read from multi files.
        """
        return data

    def read_data(self, filepath: Union[str, List[str], Tuple[str]]) -> Any:
        """
        Read data from input file.
        """
        if isinstance(filepath, (list, tuple)):
            return self.merge_input_data(self.read_single_file(x) for x in filepath)
        else:
            return self.read_single_file(filepath)

    @abc.abstractmethod
    def save_result(self, filepath: str, result: Any):
        """
        Save result to output file.
        """

    @abc.abstractmethod
    def process(self, data: Any) -> Any:
        """
        Process data and return result.
        """

    def run(self, input_path: Union[str, List[str], Tuple[str]], output_path: str = None, write_output=True):
        """
        Read from a file, and save result to another file.
        :param input_path:      file with the data
        :param output_path:     where to save the result, if not given, use input_path
        :param write_output:    whether to write the result to the output_file
        """
        data = self.read_data(input_path)
        result = self.process(data)
        if write_output:
            self.save_result(output_path or input_path, result)

class DataframeProcessor(BaseProcessor, abc.ABC):
    def __init__(self, input_dtype=None, progress=False, read_args: Dict[str, Any] = None,
                 write_args: Dict[str, Any] = None):
        self.progress = progress
        self.read_args = read_args or {}
        if input_dtype is not None:
            self.read_args['dtype'] = input_dtype
        self.write_args = write_args or {}

    def read_single_file(self, filepath: str) -> pd.DataFrame:
        return read_dataframe(filepath, **self.read_args)

    def merge_input_data(self, data: Iterable[pd.DataFrame]) -> pd.DataFrame:
        return pd.concat(data)

    def read_data(self, filepath: Union[str, List[str], Tuple[str]]) -> pd.DataFrame:
        return super().read_data(filepath)

    def save_result(self, filepath: str, result: pd.DataFrame):
        save_dataframe(filepath, result, **self.write_args)

    @abc.abstractmethod
    def process_row(self, i: Hashable, row: pd.Series) -> Optional[Dict[str, Any]]:
        """
        Process a single row of data.
        :return:    if `None`, ignore this row
        """

    def process(self, data: pd.DataFrame) -> pd.DataFrame:
        bar = data.iterrows()
        if self.progress:
            desc = "process" if self.progress is True else self.progress
            bar = tqdm.tqdm(bar, total=len(data), desc=desc)
        res = (self.process_row(i, row) for i, row in bar)
        res = (x for x in res if x is not None)
        return pd.DataFrame(res)




# -*- coding: utf-8 -*-

"""
Encapsulate methods for pandas `DataFrame`.
"""

from typing import Union, Iterable, Dict, List, Any, Sequence, Callable, Tuple, Hashable
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
import os
import pathlib
import pandas as pd
import random
import collections
from .io import ensure_parent_dir_exist
from .txt import get_file_encoding

# Compatible with different pandas versions
PD_PARAM_NEWLINE = 'lineterminator'
pd_version = [int(x) for x in pd.__version__.split('.')]
if pd_version[0] < 1 or (pd_version[0] == 1 and pd_version[1] < 5):
    PD_PARAM_NEWLINE = 'line_terminator'

FILE_FORMAT = Literal['csv', 'tsv', 'json', 'xlsx', 'parquet']
COMPRESSION_FORMAT = Literal[None, 'infer', 'snappy', 'gzip', 'brotli', 'bz2', 'zip', 'xz']

def _drop_na_values(data: Union[pd.DataFrame, Dict[str, pd.DataFrame]], axis: Literal['columns', 'rows']):
    if isinstance(data, pd.DataFrame):
        data.dropna(axis=axis, how='all', inplace=True)
    else:
        assert isinstance(data, dict)
        for df in data.values():
            df.dropna(axis=axis, how='all', inplace=True)

def _infer_file_format(file) -> str:
    if isinstance(file, str):
        return os.path.splitext(file)[1].lower()[1:]
    elif isinstance(file, pathlib.PurePath):
        suf = file.suffix
        return suf[1:] if suf.startswith('.') else suf
    elif isinstance(file, os.PathLike):
        return os.path.splitext(str(file))[1].lower()[1:]
    elif isinstance(file, pd.ExcelWriter):
        return 'xlsx'
    else:
        raise ValueError(f"Cannot infer format for type: {type(file)}")

def read_dataframe(file: str | os.PathLike, *args, sheet_name=0,
                   file_format: FILE_FORMAT = None, encoding='auto',
                   jsonl=False, dtype: type = None,
                   drop_na_columns=False, drop_na_rows=False,
                   **kwargs) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """
    read file as pandas `DataFrame`
    :param file:        the file to be read
    :param args:        extra args for `pd.read_xx()`
    :param sheet_name:      `sheet_name` for `pd.read_excel()`
    :param file_format:     csv, tsv, json ,xlsx, parquet
    :param encoding:    text file encoding
    :param jsonl:       jsonl format or not, only used in json format
    :param dtype:       `dtype` for `pd.read_xx()`
    :param drop_na_columns:     drop column if all values of the column is na
    :param drop_na_rows:        drop row if all values of the row is na
    :param kwargs:      extra kwargs for `pd.read_xx()`
    """
    # decide the file format
    if not file_format:
        file_format = _infer_file_format(file)

    for key in ['lines', 'line_delimited_json_format']:
        if key in kwargs and kwargs.pop(key):
            jsonl = True

    # handle special formats
    if file_format == 'tsv':
        # if the file format is tsv, actually same as csv
        file_format = 'csv'
        if 'sep' in kwargs:
            kwargs.pop('sep')
        kwargs['delimiter'] = '\t'
    elif file_format == 'jsonl':
        file_format = 'json'
        jsonl = True

    if encoding == 'auto' and file_format in ['csv', 'json']:
        if isinstance(file, (str, os.PathLike)):
            encoding = get_file_encoding(file, encoding=encoding)
        else:
            encoding = None

    if file_format == 'csv':
        df = pd.read_csv(file, *args, encoding=encoding, dtype=dtype, **kwargs)
    elif file_format == 'xlsx':
        df = pd.read_excel(file, *args, sheet_name=sheet_name, dtype=dtype, **kwargs)
    elif file_format == 'json':
        try:
            df = pd.read_json(file, *args, encoding=encoding, lines=jsonl, dtype=dtype, **kwargs)
        except Exception as e:
            # if failed, try again with different arg `lines`
            try:
                df = pd.read_json(file, *args, lines=not jsonl, dtype=dtype, **kwargs)
            except Exception:
                raise e
    elif file_format == 'parquet':
        df = pd.read_parquet(file, *args, **kwargs)
    else:
        raise IOError(f"Unknown file format: {file}")

    if drop_na_columns:
        _drop_na_values(df, axis='columns')
    if drop_na_rows:
        _drop_na_values(df, axis='rows')

    return df

def save_dataframe(file: Union[str | os.PathLike, 'pd.WriteBuffer[bytes]',  'pd.WriteBuffer[str]'],
                   df: Union[pd.DataFrame, Iterable[Union[pd.Series, Dict[str, Any]]]],
                   *args, sheet_name='Sheet1',
                   file_format: FILE_FORMAT = None,
                   compression: COMPRESSION_FORMAT = None,
                   index=False, index_label=None,
                   encoding='utf-8', newline='\n',
                   force_ascii=False,
                   orient='records', jsonl=True, indent=None,
                   column_mapper: Union[Dict[str, str], Sequence[str]] = None,
                   include_columns: Sequence[str] = None,
                   exclude_columns: Sequence[str] = None,
                   **kwargs):
    """
    save data into file
    :param file:                where to save the data to
    :param df:                  the data
    :param args:                extra args for df.to_xx()
    :param sheet_name:          `sheet_name` for excel format
    :param file_format:         csv, tsv, json, xlsx, parquet
    :param compression:         name of the compression to use.
                                use `None` for no compression.
    :param index:               save index or not, see docs in df.to_csv();
                                if set as str and `index_label` not set, `index_label` will be set as this
    :param index_label:         header for the index when `index` is `True`
    :param encoding:            text file encoding
    :param newline:             text file newline
    :param force_ascii:         `force_ascii` for json format
    :param orient:              `orient` for json format
    :param jsonl:               jsonl format or not
    :param indent:              indent for json format
    :param column_mapper:       rename columns; if set, columns not list here will be ignored
    :param include_columns:     if set, columns not list here will be ignored
    :param exclude_columns:     if set, columns list here will be ignored
    :param kwargs:              extra kwargs for df.to_xx()
    """
    # decide file format
    if not file_format:
        file_format = _infer_file_format(file)

    # convert data to be a dataframe
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)

    for key in ['lines', 'line_delimited_json_format']:
        if key in kwargs and kwargs.pop(key):
            jsonl = True

    # deal with columns
    if column_mapper:
        df = df.rename(columns=column_mapper)
    if exclude_columns:
        df = df.drop(exclude_columns, axis=1)
    if include_columns:
        df = df.reindex(include_columns, axis=1)

    # ensure parent dir exists
    if isinstance(file, (str, os.PathLike)):
        ensure_parent_dir_exist(file)

    # compatible for set index just use arg `index`
    if index_label is None and isinstance(index, str):
        index, index_label = True, index

    # handle special formats
    if file_format == 'tsv':
        # tsv is actually a csv
        file_format = 'csv'
        kwargs['sep'] = '\t'
    elif file_format == 'jsonl':
        file_format = 'json'
        jsonl = True

    # save to file for different format
    if file_format == 'csv':
        kwargs[PD_PARAM_NEWLINE] = newline
        df.to_csv(file, *args, compression=compression, index=index, index_label=index_label,
                  encoding=encoding, **kwargs)
    elif file_format == 'xlsx':
        df.to_excel(file, *args, index=index, index_label=index_label, sheet_name=sheet_name, **kwargs)
    elif file_format == 'json':
        if jsonl:
            orient = 'records'
            indent = None
        if orient not in ['split', 'table']:
            index = True
        df.to_json(file, *args, compression=compression, index=index,
                   force_ascii=force_ascii, orient=orient, lines=jsonl,
                   indent=indent, **kwargs)
    elif file_format == 'parquet':
        df.to_parquet(file, *args, compression=compression, index=index, **kwargs)
    else:
        raise IOError(f"Unknown file format: {file}")

def iter_dataframe(data: pd.DataFrame,
                   progress_bar: Union[bool, str, 'tqdm', Callable[[Iterable[Any]], 'tqdm']] = False
                   ) -> Iterable[Tuple[Hashable, pd.Series]]:
    """
    iter dataframe rows, may show a progress bar
    :param data:            dataframe
    :param progress_bar:    show a progress bar or not
                            if set a non-empty string, the string will be set as the progress bar description
    """
    rows = data.iterrows()
    if progress_bar:
        from tqdm import tqdm
        if isinstance(progress_bar, tqdm):
            progress_bar.iterable = rows
            rows = progress_bar
        elif isinstance(progress_bar, str):
            rows = tqdm(rows, total=len(data), desc=progress_bar)
        elif callable(progress_bar):
            rows = progress_bar(rows)
        else:
            rows = tqdm(rows, total=len(data))
    return rows

def extract_dataframe_sample(data: pd.DataFrame,
                             filter_func: Callable[[pd.Series], bool],
                             size=0, shuffle=False,
                             return_format: Literal['df', 'dataframe', 'list'] = 'dataframe',
                             progress_bar=False) -> Union[pd.DataFrame, List[pd.Series]]:
    """
    extract sample from a dataframe
    :param data:            original data
    :param filter_func:     bool function, `True` means to reserve the row
    :param size:            max size for the result
    :param shuffle:         shuffle result or not
    :param progress_bar:    passed to `iter_dataframe()`
    :param return_format:   one of {'dataframe', 'list'}
    """
    result = [row for _, row in iter_dataframe(data, progress_bar=progress_bar) if filter_func(row)]
    if shuffle:
        random.shuffle(result)
    if 0 < size < len(result):
        result = result[:size]
    if return_format == 'df' or return_format == 'dataframe':
        try:
            return pd.DataFrame(result)
        except pd.errors.InvalidIndexError:
            return pd.DataFrame([{k: v for k, v in x.items()} for x in result])
    elif return_format == 'list':
        return result
    raise ValueError("Param 'return_format' should be one of {'dataframe', 'list'}.")

def is_empty_text(s: str) -> bool:
    return pd.isna(s) or not s

def is_nonempty_text(s: str) -> bool:
    return pd.notna(s) and isinstance(s, str) and s

def is_blank_text(s: str) -> bool:
    return pd.isna(s) or isinstance(s, str) and not s.strip()

def is_non_blank_text(s: str) -> bool:
    return pd.notna(s) and isinstance(s, str) and s.strip()

def join_values(values: Sequence[Any], sep=None) -> str:
    if not values:
        return ''
    if len(values) == 1:
        return str(values[0])
    return sep.join(map(str, values)) if sep else values

def merge_dataframe_rows(data: pd.DataFrame, col_id='ID', na=None, join_sep=None, progress_bar=False) -> pd.DataFrame:
    """
    merge rows of same id to one row, similar to group by in sql
    :param data:            original data
    :param col_id:          column name for the id col
    :param na:              values to be treated as na
    :param join_sep:        seperator to join multi values
    :param progress_bar:    passed to `iter_dataframe()`
    """
    if na is None:
        na = set()
    elif isinstance(na, str):
        na = {na}
    else:
        na = set(na)
    counts = collections.defaultdict(lambda: collections.defaultdict(collections.Counter))
    rows = iter_dataframe(data, progress_bar=progress_bar)
    for i, row in rows:
        eid = row[col_id]
        for k, v in row.items():
            if pd.notna(v) and v not in na:
                counts[eid][k][v] += 1
    result = []
    for x in counts.values():
        item = {col: join_values(list(values.keys()), sep=join_sep) for col, values in x.items()}
        result.append(item)
    return pd.DataFrame(result)



# -*- coding: utf-8 -*-

from typing import Any, Callable

def join_values(*values: Any, sep='', func: Callable[[Any], str] = str, do_trim=False, ignore_empty=False):
    def f():
        for x in values:
            s = func(x)
            if do_trim:
                s = s.strip()
            if s or not ignore_empty:
                yield s
    return sep.join(f())
